"""
The TASKS file is a text file containing one task per line:

RUN 4.7.2
RUN 666 -b "time_commit"

The format is the following: `TASK_TYPE ARGUMENTS`.
The supported TASK_TYPE are:
    * RUN: launch asv on the ARGUMENTS. Accepts all arguments that asv run
      accepts.

Once a TASK is done, the scheduler will remove it from the the task file. You
can edit this file, please use an editor that detects when the tasks file
changed since you start editing it (at least nano does it).
"""
import re
import os
import argparse
import sys
import time
import shlex
import subprocess
import logging

from os.path import dirname, join, abspath

DEFAULT_RUN_SCRIPT = abspath(join(dirname(__file__), "main-run.sh"))

RUN_SCRIPT = os.environ.get("SCHEDULER_TEST_RUN_SCRIPT", DEFAULT_RUN_SCRIPT)

DEFAULT_BISECT_SCRIPT = abspath(join(dirname(__file__), "main-bisect.sh"))

BISECT_SCRIPT = os.environ.get("SCHEDULER_TEST_BISECT_SCRIPT", DEFAULT_BISECT_SCRIPT)

LOGGER = logging.getLogger(__name__)
LOG_FORMAT = "[%(asctime)s] %(message)s"


def read_next_scheduling(task_file_path):
    with open(task_file_path, "r") as f:
        first_line = f.readline().strip()

    return first_line


def run(task):
    splitted = shlex.split(task)
    args = [RUN_SCRIPT] + splitted

    try:
        print("RUNNING %r" % args)
        subprocess.check_call(args)
    except subprocess.CalledProcessError as e:
        msg = "Task %r failed with error code %r"
        LOGGER.error(msg, task, e.returncode)
        print("\n" + msg % (task, e.returncode))
        # XXX find a way to report an error without removing a task

def bisect(task):
    # We need to parse the task arguments not in POSIX mode because POSIX mode
    # swallow single quote inside values
    splitted = shlex.split(task, posix=False)

    # Assume the test name is the last argument. We need to escape it as it's
    # expected to be a regex but the full test name contains non-alphanumeric
    # characters (at least `(` and `)`)
    splitted[-1] = splitted[-1].replace('(', '\(').replace(')', '\)')

    args = [BISECT_SCRIPT] + splitted

    try:
        print("Bisecting %r" % args)
        subprocess.check_call(args)
    except subprocess.CalledProcessError as e:
        msg = "Task %r failed with error code %r"
        LOGGER.error(msg, task, e.returncode)
        print("\n" + msg % (task, e.returncode))
        # XXX find a way to report an error without removing a task


def remove_task(task, task_file_path):
    """ Remove the first line of the scheduling TASK if it match passed task
    """
    with open(task_file_path, "r") as f:
        lines = f.readlines()

    if lines[0].strip() == task:
        lines = lines[1:]

        # XXX We should instead write to a temporary file and move it
        with open(task_file_path, "w") as f:
            f.writelines(lines)

        print("Removed task %r" % task)
    else:
        print("Don't remove non-matching first task %r" % task)


def main(task_file_path):
    while True:
        next_task = read_next_scheduling(task_file_path)

        if next_task:

            # Try splitting it
            splitted = next_task.split(" ", 1)
            assert len(splitted) > 0
            task_type = splitted[0]

            task_arguments = None
            if len(splitted) == 2:
                task_arguments = splitted[1]

            print("=" * 80)
            print("Got task %r %r" % (task_type, task_arguments))
            LOGGER.info("Starting %r %r" % (task_type, task_arguments))

            try:
                if task_type == "RUN":
                    run(task_arguments)
                elif task_type == "BISECT":
                    bisect(task_arguments)
                elif task_type == "EXIT":
                    return
                else:
                    msg = "Unknown task %s"
                    print(msg % task_type)
            finally:
                LOGGER.info("Ending %r %r", task_type, task_arguments)
                print("\n" * 2)
                remove_task(next_task, task_file_path)
                print("=" * 80)

        time.sleep(1)


def setup_logging(logging_file):
    # Reset handlers
    LOGGER.handlers = []

    LOGGER.setLevel(logging.INFO)

    # Add handler
    console = logging.FileHandler(logging_file)

    console.setLevel(logging.INFO)

    formatter = logging.Formatter(LOG_FORMAT)
    console.setFormatter(formatter)

    LOGGER.addHandler(console)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--log", help="The logging file to use")
    parser.add_argument("TASKS", help="The file path to the task file")

    args = parser.parse_args(sys.argv[1:])

    # Configure logging
    if args.log:
        setup_logging(args.log)

    main(args.TASKS)
